// Calculation of NRG chain coefficients
// 
// Rok Zitko, rok.zitko@ijs.si, March 2009
// $Id: nrgchain.cc,v 1.3 2009/03/20 13:38:32 rok Exp rok $

/*
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 * 
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 * 
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 * */

#include <iostream>
#include <fstream>
#include <iomanip>
#include <cmath>
#include <cstdlib>
#include <cassert>
#include <cfloat>
#include <utility>
#include <vector>
#include <string>
#include <map>
#include <ctime>

#include <gmp.h>

using namespace std;

#include "lambda.h"
#include "linint.h"
#include "io.h"
#include "parser.h"
#include "load.h"
#include "calc.h"

string param_fn = "param"; // file with input parameters
LAMBDA Lambda; // discretization parameter
double z; // twist parameter
double xmax; // higher boundary of the x=j+z interval, where the ODE
             // was solved numerically
unsigned int mMAX; // the number of coefficients computed (max index)
unsigned int Nmax; // the length of the Wilson chain (max index)

unsigned int preccpp; // precision for GMP

Vec vecrho_pos, vecrho_neg; // rho, for positive and negative energies
LinInt rho_pos, rho_neg;
IntLinInt intrho_pos, intrho_neg; // integrated rho
LinInt g_pos, g_neg; // g(x)
LinInt f_pos, f_neg;

typedef vector<double> Table;

Table de_pos, de_neg, du_pos, du_neg;

bool adapt; // If adapt=false --> g(x)=1.

// eps(x) = D g(x) Lambda^(2-x) for x>2.
double eps_pos(double x)
{
  const double gx = (adapt ? g_pos(x) : 1.0);
  return (x <= 2.0 ? 1.0 : gx * Lambda.power(2.0-x));
}

double eps_neg(double x)
{
  const double gx = (adapt ? g_neg(x) : 1.0);
  return (x <= 2.0 ? 1.0 : gx * Lambda.power(2.0-x));
}

// Eps(x) = D f(x) Lambda^(2-x)
inline double Eps_pos(double x)
{
  assert(x >= 1);
  return f_pos(x) * Lambda.power(2.0-x);
}

inline double Eps_neg(double x)
{
  assert(x >= 1);
  return f_neg(x) * Lambda.power(2.0-x);
}

void about(ostream &F = cout)
{
   F << "# Calculation of NRG coefficients" << endl;
   F << "# Rok Zitko, rok.zitko@ijs.si, March 2009" << endl;
   F << "# $Id: nrgchain.cc,v 1.3 2009/03/20 13:38:32 rok Exp rok $" << endl;
}

void cmd_line(int argc, char *argv[])
{
   if (argc > 2) {
      cerr << "Usage: nrgchain [P|N] [param_filename]" << endl;
      exit(1);
   }
   
   if (argc == 2) {
      param_fn = string(argv[1]);
   }
}

void set_parameters()
{
  cout << setprecision(PREC);

  Lambda = LAMBDA(P("Lambda", 2.0));
  assert(Lambda > 1.0);

  z = P("z", 1.0);
  assert(0 < z && z <= 1.0); 

  adapt = Pbool("adapt", false); // Enable adaptable g(x)? Default is false!!

  xmax = P("xmax", 30); // Interval [1..xmax]
  assert(xmax >= 1.0);

  mMAX = Pint("mMAX", 0); // Maximal index of coefficients (e,f)
  Nmax = Pint("Nmax", 0); // Maximal site index in the Wilson chain

  preccpp = Pint("preccpp", 2000); // Precision for GMP
  assert(preccpp > 10);

  cout << "# Lambda=" << Lambda;
  cout << " z=" << z << endl;
  cout << "# xmax=" << xmax;
  cout << " mMAX=" << mMAX;
  cout << " Nmax=" << Nmax << endl;
}

void load_rho()
{
  const string rhofn = Pstr("dos", "Delta.dat");
  vecrho_pos = load_rho(rhofn, POS);
  rho_pos = LinInt(vecrho_pos);
  vecrho_neg = load_rho(rhofn, NEG);
  rho_neg = LinInt(vecrho_neg);

  Vec vecintrho_pos(vecrho_pos);
  integrate(vecintrho_pos);
  intrho_pos = IntLinInt(vecrho_pos, vecintrho_pos);

  Vec vecintrho_neg(vecrho_neg);
  integrate(vecintrho_neg);
  intrho_neg = IntLinInt(vecrho_neg, vecintrho_neg);
}

void load_g()
{
  const string gfn_pos = "GSOL.dat";
  Vec vecg_pos = load_g(gfn_pos);
  g_pos = LinInt(vecg_pos);

  const string gfn_neg = "GSOLNEG.dat";
  Vec vecg_neg = load_g(gfn_neg);
  g_neg = LinInt(vecg_neg);
}

void load_f()
{
  const string ffn_pos = "FSOL.dat";
  Vec vecf_pos = load_g(ffn_pos); // same load_g() function as for g
  f_pos = LinInt(vecf_pos);

  const string ffn_neg = "FSOLNEG.dat";
  Vec vecf_neg = load_g(ffn_neg);
  f_neg = LinInt(vecf_neg);
}

// The factor that multiplies eigenvalues of the Wilson chain Hamiltonian
// in order to obtain the eigenvalues of the true Hamiltonian (at scale D).
double SCALE(int N)
{
  return (1.0 - 1./Lambda)/log(Lambda) * 
    pow(Lambda, -(N-1.0)/2.0 + 1.0 - z);
}

inline double sqr(double x) { return x*x; }

void tables()
{
  const double int_pos = integrate_ab(vecrho_pos, 0.0, 1.0);
  const double int_neg = integrate_ab(vecrho_neg, 0.0, 1.0);
  const double theta = int_pos + int_neg;
  cout << "# int_pos=" << int_pos << " int_neg=" << int_neg 
       << " theta=" << theta << endl;

  Table df_pos(mMAX+1), df_neg(mMAX+1);
  Table du0_neg(mMAX+1), du0_pos(mMAX+1);

  de_pos.resize(mMAX+1);
  de_neg.resize(mMAX+1);

  //  double zfaktor = Lambda.factor();

  for (unsigned int m = 0; m <= mMAX; m++) {
    df_pos[m] = intrho_pos(eps_pos(z+m+1)) - intrho_pos(eps_pos(z+m+2));
    df_neg[m] = intrho_neg(eps_neg(z+m+1)) - intrho_neg(eps_neg(z+m+2));

    du0_pos[m] = sqrt(df_pos[m])/sqrt(theta);
    du0_neg[m] = sqrt(df_neg[m])/sqrt(theta);

    de_pos[m] = Eps_pos(z+m+1);
    de_neg[m] = Eps_neg(z+m+1);
  }

  double checksum = 0.0;
  for (unsigned int m = 0; m <= mMAX; m++) 
    checksum += sqr(du0_pos[m])+sqr(du0_neg[m]);
  
  cout << "# 1-checksum=" << 1-checksum << endl;

  du_pos.resize(mMAX+1);
  du_neg.resize(mMAX+1);

  for (unsigned int m = 0; m <= mMAX; m++) {
    du_pos[m] = du0_pos[m]/sqrt(checksum);
    du_neg[m] = du0_neg[m]/sqrt(checksum);
  }

  for (unsigned int m = 0; m <= mMAX; m++) {
    cout << "# " << m << " " 
         << du_pos[m] << " " << du_neg[m] << " "
         << de_pos[m] << " " << de_neg[m] << endl;
  }
}

class my_mpf
{
private:
  mpf_t val;

public:
  my_mpf() { mpf_init(val); }
  ~my_mpf() { mpf_clear(val); }
  inline operator mpf_t &() { return val; }
};

// Fix normalization of u_{n,m}, v_{n,m} to 1. IMPORTANT: pass by
// reference!
void fix_norm(my_mpf up[], my_mpf um[], unsigned int mMAX)
{
  // Constants
  my_mpf mpZERO, mpONE;
  mpf_set_str(mpONE, "1.e0", 10);

  my_mpf sum, temp, tempsq;

  mpf_set(sum, mpZERO);
  for (unsigned int m = 0; m <= mMAX; m++) {
    mpf_mul(tempsq, up[m], up[m]);
    mpf_add(sum, sum, tempsq);
    mpf_mul(tempsq, um[m], um[m]);
    mpf_add(sum, sum, tempsq);
  }
  mpf_sqrt(temp, sum);

  for (unsigned int m = 0; m <= mMAX; m++) {
    mpf_div(up[m], up[m], temp);
    mpf_div(um[m], um[m], temp);
  }
}

#define HIGHPREC(val) setw(30) << setprecision(16) << (val) \
                               << setprecision(PREC)

void tridiag()
{
  ofstream XI, ZETA;
  safe_open(XI, "xi.dat"); // hopping constants 
  safe_open(ZETA, "zeta.dat"); // on-site energies
   
  mpf_set_default_prec(preccpp);
  cout << "Using precision of " << preccpp << " digits." << endl;

  // Constants
  my_mpf mpZERO;

  // Temporary MP variables
  my_mpf temp, tempsq, sum;

  my_mpf mpxi; // xi
  my_mpf xi2; // xi^2
  my_mpf mpzeta; // zeta

  my_mpf xi_prev, xi2_prev; // values in previous iteration

  my_mpf up[mMAX+1];
  my_mpf up_prev[mMAX+1];
  my_mpf up_prev2[mMAX+1];
  my_mpf um[mMAX+1];
  my_mpf um_prev[mMAX+1];
  my_mpf um_prev2[mMAX+1];
  my_mpf ep1[mMAX+1];
  my_mpf em1[mMAX+1];
  my_mpf ep2[mMAX+1];
  my_mpf em2[mMAX+1];
  for (unsigned int m = 0; m <= mMAX; m++) {
    mpf_set_d(up_prev[m], du_pos[m]);
    mpf_set_d(um_prev[m], du_neg[m]);
    mpf_set_d(ep1[m], de_pos[m]);
    mpf_set_d(em1[m], de_neg[m]);
    mpf_mul(ep2[m], ep1[m], ep1[m]);
    mpf_mul(em2[m], em1[m], em1[m]);
  }

  fix_norm(up_prev, um_prev, mMAX);

  for (unsigned int n = 0; n <= Nmax; n++) {
    // Calculate zeta_n, xi2_n and xi_n
    mpf_set(mpzeta, mpZERO);
    mpf_set(xi2, mpZERO);
    for (unsigned int m = 0; m <= mMAX; m++) {
      // up_prev = u^+_{n,m}
      mpf_mul(tempsq, up_prev[m], up_prev[m]);
      mpf_mul(temp, tempsq, ep2[m]);
      mpf_add(xi2, xi2, temp);
      mpf_mul(temp, tempsq, ep1[m]);
      mpf_add(mpzeta, mpzeta, temp);

      // um_prev = u^-_{n,m}
      mpf_mul(tempsq, um_prev[m], um_prev[m]);
      mpf_mul(temp, tempsq, em2[m]);
      mpf_add(xi2, xi2, temp);
      mpf_mul(temp, tempsq, em1[m]);
      mpf_sub(mpzeta, mpzeta, temp);
    }

    // subtract xi^2_{n-1}
    mpf_sub(xi2, xi2, xi2_prev);

    // subtract zeta^2_n
    mpf_mul(tempsq, mpzeta, mpzeta);
    mpf_sub(xi2, xi2, tempsq);

    mpf_sqrt(mpxi, xi2);

    // compute u_{n+1,m}, v_{n+1,m}
    for (unsigned int m = 0; m <= mMAX; m++) {
      // zeta=zeta_n
      mpf_sub(temp, ep1[m], mpzeta);
      // up_prev[m]=u_{n,m}
      mpf_mul(up[m], temp, up_prev[m]);

      // xi_prev=xi_{n-1}, up_prev2=u_{n-1,m}
      mpf_mul(temp, xi_prev, up_prev2[m]);
      mpf_sub(up[m], up[m], temp);

      // xi=xi_n
      mpf_div(up[m], up[m], mpxi);

      mpf_neg(temp, em1[m]);
      mpf_sub(temp, temp, mpzeta);
      mpf_mul(um[m], temp, um_prev[m]);
      mpf_mul(temp, xi_prev, um_prev2[m]);
      mpf_sub(um[m], um[m], temp);
      mpf_div(um[m], um[m], mpxi);
    }

    fix_norm(up, um, mMAX);

    // Recalculate xi, xi2
    mpf_set(sum, mpZERO);
    for (unsigned int m = 0; m <= mMAX; m++) {
      mpf_mul(temp, up[m], up_prev[m]);
      mpf_mul(temp, temp, ep1[m]);
      mpf_add(sum, sum, temp);
      mpf_mul(temp, um[m], um_prev[m]);
      mpf_mul(temp, temp, em1[m]);
      mpf_sub(sum, sum, temp);
    }
    mpf_set(mpxi, sum);
    mpf_mul(xi2, mpxi, mpxi);

    // Save results
    double dxi = mpf_get_d(mpxi);
    double dzeta = mpf_get_d(mpzeta);
    double coef_xi = dxi/SCALE(n+1);
    double coef_zeta = dzeta; // NOT RESCALED!!!

    XI << coef_xi << endl;
    ZETA << coef_zeta << endl;

    cout << "  xi(" << n << ")=" << HIGHPREC( dxi )
         << " --> " << HIGHPREC( coef_xi )<< endl;
    cout << "zeta(" << n << ")=" << HIGHPREC( dzeta ) << endl;

    // Store results from previous iteration
    mpf_set(xi_prev, mpxi);
    mpf_set(xi2_prev, xi2);

    for (unsigned int m = 0; m <= mMAX; m++) {
      mpf_set(um_prev2[m], um_prev[m]);
      mpf_set(up_prev2[m], up_prev[m]);
      mpf_set(um_prev[m], um[m]);
      mpf_set(up_prev[m], up[m]);
    }
  }
}

int main(int argc, char *argv[])
{
   clock_t start_clock = clock();
   
   about();
   cmd_line(argc, argv);
   parser(param_fn);
   set_parameters();

   load_rho();
   if (adapt) {
     load_g();
   }
   load_f();

   tables();
   tridiag();
   
   clock_t end_clock = clock();
   cout << "# Elapsed " << double(end_clock-start_clock)/CLOCKS_PER_SEC << " s"
     << endl;
}
